/* $Id: array.c,v 1.9 1997/03/28 14:36:54 queinnec Exp $ */

#include "config.h"

#if HAVE_STRING_H
#include <string.h>
#else
int strcmp (const char *, const char *);
char *strcpy (char *, const char *);
int strlen (const char *);
#endif

#include <assert.h>
#include <stdio.h>
#include <ctype.h>

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <sys/types.h>
#if HAVE_MMAP
#include <sys/mman.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

#if HAVE_LIBZ
#include <zlib.h>
#endif

#include "constants.h"
#include "array.h"
#include "completion.h"
#include "translation.h"

static const char *rcsid = "$Id: array.c,v 1.9 1997/03/28 14:36:54 queinnec Exp $";

/* I don't support any more without ECONOMICALLOC: save and load would
   become too complex if they need to handle both cases (and would be
   very inefficient if !ECONOMICALLOC) */
#define ECONOMICALLOC 1

/*
 * Danger: any change in this structure implies a change in the save format
 */
struct Array {
    
    unsigned char translations[256];
    
    int  sorted;
    unsigned int  nballocated;
    unsigned int  nbused;
    unsigned char **data;

    int leftinpool;             /* a local pool is used if savable is TRUE, */
    unsigned char *pool;        /* else a common pool is used. */

    int savable;
    unsigned int nbstringblocks;
    unsigned char **blocks;
};

/* Used when resizing the arrays data or blocks */
#define ALLOCBUNCH 100




/* If len == 0, it's computed here. */
static unsigned char *str_dup (const unsigned char *s, unsigned int len)
{
    unsigned char *r;
    if (len == 0)
      len = strlen (s);
    r = malloc (len + 1);
    if (r == NULL) {
        fprintf (stderr, "array: str_dup: malloc failed\n");
        return NULL;
    }
    strcpy (r, s);
    return r;
}



/* ATTENTION:
 * keepstring returns a pointer which must not be passed to free.
 */

#if ECONOMICALLOC

/* Make it a power of 2 minus what is needed for malloc internal maintenance
   (generally one pointer, rarely more).
   Here 32Ko - 32 bytes
 */
/* #define ALLOCPOOL 8160 */
#define ALLOCPOOL 32736

static unsigned char *keepstring (Array *a, const unsigned char *s,
                                  unsigned int len)
{
    static unsigned int commonleftinpool = 0;
    static unsigned char *commonpool = NULL;
    unsigned int leftinpool;
    unsigned char *pool;
    unsigned char *r;

    if (len >= ALLOCPOOL) {      /* be paranoiac! */
        r = str_dup (s, len);
        return r;
    }
    if (a->savable) {
        pool = a->pool; leftinpool = a->leftinpool;
    } else {
        pool = commonpool; leftinpool = commonleftinpool;
    }

    if (len >= leftinpool) {
        /* Allocate a new pool. */
        pool = malloc (ALLOCPOOL);
        if (pool == NULL) {
            fprintf (stderr, "array: dupstring: malloc failed\n");
            return NULL;
        }
        leftinpool = ALLOCPOOL;
        if (a->savable) {
            /* Record the new pool in the list of allocated pools. */
            if (a->nbstringblocks == 0)
              a->blocks = malloc (ALLOCBUNCH * sizeof(char*));
            else if (a->nbstringblocks % ALLOCBUNCH == 0)
              a->blocks = realloc (a->blocks,
                                   (a->nbstringblocks + ALLOCBUNCH) * sizeof(char*));
            a->blocks[a->nbstringblocks] = pool;
            a->nbstringblocks++;
        }
    }

    r = pool;
    strcpy (pool, s);
    pool += len;
    pool++;
    leftinpool -= len;
    leftinpool--;

    if (a->savable) {
        a->pool = pool; a->leftinpool = leftinpool;
    } else {
        commonpool = pool; commonleftinpool = leftinpool;
    }
    return r;
}

#define freestring(s) /* empty */

#else /* !ECONOMICALLOC */

#define keepstring(a,s,len) (str_dup ((s), (len)))
#define freestring(s) (free ((s)))

#endif /* !ECONOMICALLOC */

Array *array_create (int savable, const unsigned char *translation_table)
{
    Array *r;
    unsigned int i;
    r = malloc (sizeof (Array));
    if (r == NULL) {
        fprintf (stderr, "Array: create: malloc failed\n");
        return NULL;
    }

    if (translation_table == NULL) {
        for (i = 0; i < 256; i++)
          r->translations[i] = i;
    } else {
        for (i = 0; i < 256; i++)
          r->translations[i] = translation_table[i];
    }
    
    r->sorted = FALSE;
    
    r->nbused = 0;
    r->nballocated = ALLOCBUNCH;
    r->data = malloc (ALLOCBUNCH * sizeof (char *));
    if (r->data == NULL) {
        free (r);
        fprintf (stderr, "Array: create: malloc failed\n");
        return NULL;
    }
    for (i = 0; i < r->nballocated; i++) {
        r->data[i] = NULL;
    }

    r->leftinpool = 0;
    r->pool = NULL;

    r->savable = savable;
    r->nbstringblocks = 0;
    r->blocks = NULL;
    return r;
}

/*
 * Array destruction.
 */
int   array_destroy (Array *array)
{
    unsigned int i;
    assert (array != NULL);
    if (! array->savable) {
        fprintf (stderr, "array: destroy: memory cannot be freed\n");
    } else {
        for (i = 0; i < array->nbstringblocks; i++)
          free (array->blocks[i]);
        free (array->blocks);
    }
    free (array->data);
    free (array);
    return OK;
}

#if 0
/*
 * Convert word to lowercase.
 * Result is put into a static zone (which can hold any number of chars)
 * If len is 0, it is computed by convertlowcase.
 */
static unsigned char *convertlowcase (const unsigned char *word, unsigned int len)
{
    static unsigned char *lowcase = NULL;
    static unsigned int lowcasesize;
    unsigned char *to;
    if (len == 0)
      len = strlen (word);
    if (lowcase == NULL) {
        lowcase = malloc (len + 1);
        lowcasesize = len;
    } else if (lowcasesize < len) {
        /* increase the "static" zone if necessary */
        lowcase = realloc (lowcase, len + 1);
        lowcasesize = len;
    }
    if (lowcase == NULL) {
        fprintf (stderr, "array: convert lower case: malloc failed\n");
        return NULL;
    }
    to = lowcase;
    for ( ; *word != 0; word++) {
        *to++ = (isupper (*word) ? tolower (*word) : *word);
    }
    *to = 0;
    return lowcase;
}
#endif /* 0 */

/*
 * Word insertion.
 */
int   array_insert (Array *array, const unsigned char *word)
{
    unsigned char *p;
    unsigned int len;
    assert (array != NULL);
    len = strlen (word);
    if (len == 0)
      return ERROR;

    if (array->nbused == array->nballocated) {
        unsigned int i;
        array->nballocated += ALLOCBUNCH;
        array->data = realloc(array->data, array->nballocated * sizeof(char*));
        if (array->data == NULL) {
            fprintf (stderr, "array: insert: realloc failed\n");
            return ERROR;
        }
        for (i = array->nbused; i < array->nballocated; i++) {
            array->data[i] = NULL;
        }
    }
    p = keepstring (array, word, len);
    if (p == NULL) {
        fprintf (stderr, "array: insert: maalloc failed\n");
        return ERROR;
    }
    array->data[array->nbused] = p;
    array->nbused++;
    array->sorted = FALSE;
    return OK;
}

/*
 * Search *an* index where "word" matches up to "len"'th character.
 * If "len" is 0, "word" must match exactly.
 * This use a dichotomic search.
 */
static int find_index (Array *a, const unsigned char *word, unsigned int len)
{
    unsigned int low;
    unsigned int high;
    unsigned int middle;
    if (a->nbused == 0)
      return -1;
    if (len == 0)
      len = 65535; /* well, some big int like UINT_MAX. Why not (unsigned)-1? */
    low = 0;
    high = a->nbused - 1;
    if (tr_strncmp (word, a->data[low], len, a->translations) < 0)
      return -1;
    else if (tr_strncmp (word, a->data[high], len, a->translations) > 0)
      return -1;
    while (low < high) {
        int cmp;
        middle = (low + high) / 2;
        cmp = tr_strncmp (a->data[middle], word, len, a->translations);
        if (cmp < 0)
          low = middle + 1;
        else if (cmp == 0)
          return middle;
        else
          high = middle - 1;
    }
    if (tr_strncmp (a->data[low], word, len, a->translations) == 0)
      return low;
    else
      return -1;
}

/*
 * Word removal.
 */
int   array_delete (Array *array, const unsigned char *word)
{
    int index;
    unsigned int i;
    index = find_index (array, word, 0);
    if (index == -1)
      return (ERROR);
    /* No message if ECONOMICALLOC, would be to talkative */
    /* freestring (array->data[index]); */
    for (i = index+1; i < array->nbused; i++) {
        array->data[i-1] = array->data[i];
    }
    array->nbused--;
    return OK;
}

/*
 * Checking for word presence.
 */
int   array_check (Array *array, const unsigned char *word)
{
    assert (array->sorted);
    if (find_index (array, word, 0) == -1)
      return FALSE;
    else
      return TRUE;
}

unsigned int array_nb_of_words (const Array *a)
{
    return a->nbused;
}

unsigned char *array_get_translations (Array *a)
{
    return a->translations;
}



static unsigned char *current_translations = NULL;

static int cmpword (const void *a, const void *b)
{
    const unsigned char *mota = *(const unsigned char * const *)a;
    const unsigned char *motb = *(const unsigned char * const *)b;
    return (tr_strcmp (mota, motb, current_translations));
}

int array_endinsertion (Array *a)
{
    if (a->sorted) {
        return OK;
    }
    if (a->nbused != 0) {
        current_translations = a->translations;
        qsort (a->data, a->nbused, sizeof(char *), cmpword);
        /* What follows looks like a good idea, but it is useless for
         * imoviedb as libreadline removes duplicates in completion.
         * Moreover, with ECONOMICALLOC, we can't free the data, and what
         * is precious is memory and time. Here we are losing time without
         * saving memory. What should be done is avoiding to insert the
         * word in the first place, but it's rather expensive.
         * Moreover, the algorithm is rather stupid as it's compacting the
         * array every times a duplicate is found. It should instead replace
         * the duplicate with a dead_slot, and compacting the array once
         * at the end.*/
#if 0
        /* Remove duplicates if necessary */
        {
            unsigned int i, j;
            /* Search the first duplicate word */
            for (i = 1; i < a->nbused; i++) {
                if (tr_strcmp (a->data[i], a->data[i-1], a->translations) == 0)
                  break;
            }
            /* If found... */
            if (i != a->nbused) {
                /* Move back the names which are not duplicate of the
                 *  previously copied word. */
                for (j = i; j < a->nbused; j++) {
                    if (tr_strcmp (a->data[j], a->data[i-1], a->translations) != 0) {
                        a->data[i++] = a->data[j];
                    } else {
                        freestring (a->data[j]);
                    }
                }
                a->nbused = i;
            }
        }
#endif /* 0 */
    }
    a->sorted = TRUE;
    return OK;
}


/********************* COMPLETION ********************/

struct ArrayCompletion {
    Array            *array;
    unsigned char    *prefix;
    unsigned int      len;      /* len of prefix */
    int               index;    /* current index in array, -1 if invalid */
    int               number;   /* of completions (-1 if not yet computed) */
    Compl_Success     success;
};

ArrayCompletion *array_createcompletion (Array *a)
{
    ArrayCompletion *r;
    r = (ArrayCompletion *) malloc (sizeof (ArrayCompletion));
    if (r == NULL) {
        fprintf (stderr, "Array: create completion: malloc failed\n");
        return NULL;
    }
    r->array = a;
    r->prefix = NULL;
    r->index = -1;
    r->len = 0;
    r->number = -1;
    r->success = COMPL_NOTDONE;
    return r;
}

int array_destroycompletion (ArrayCompletion *ac)
{
    if (ac == NULL)
      return ERROR;
    if (ac->prefix != NULL)
      free (ac->prefix);
    free (ac);
    return OK;
}

Compl_Success array_trycompletion (ArrayCompletion *ac, const unsigned char *word)
{
    int index;
    Array *a;
    assert (ac != NULL);
    assert (ac->array->sorted);
    if (*word == 0) {
        ac->success = COMPL_FAIL;
        return ac->success;
    }
    if (ac->prefix != NULL)
      free (ac->prefix);
    ac->len = strlen (word);
    ac->index = -1;
    ac->number = -1;
    a = ac->array;
    ac->prefix = str_dup (word, ac->len);
    if (ac->prefix == NULL) {
        ac->success = COMPL_FAIL;
        return ac->success;
    }
    index = find_index (a, word, ac->len);
    if (index == -1) {
        ac->success = COMPL_FAIL;
    } else {
        /* move back to the first name which matches */
        if (index > 0) {
            index--;
            while ((index >= 0) &&
                   (tr_strncmp(a->data[index],word,ac->len,a->translations) == 0)) {
                index--;
            }
            index++;
        }
        ac->index = index;

        if (tr_strcmp (a->data[index], word, a->translations) == 0) {
            ac->success = COMPL_EXACT;
            if ((index+1) < a->nbused) {
                if (tr_strncmp(a->data[index+1],word,ac->len,a->translations) == 0)
                  ac->success = COMPL_EXACTBUTLONGER;
            }
        } else {
            ac->success = COMPL_PARTIAL;
        }
    } /* index != -1 */
    return ac->success;
}

int array_nbofcompletions (ArrayCompletion *ac)
{
    Array *a;
    unsigned int index;
    unsigned int num = 0;
    assert (ac != NULL);
    if (ac->success == COMPL_NOTDONE)
      return -1;
    if (ac->number != -1)
      return ac->number;
    if (ac->success == COMPL_FAIL) {
        ac->number = 0;
        return ac->number;
    }
    a = ac->array;
    assert (ac->index != -1);
    index = ac->index;
    while ((index < a->nbused) &&
           (tr_strncmp(a->data[index],ac->prefix,ac->len,a->translations) == 0)) {
        index++;
        num++;
    }
    ac->number = num;
    return num;
}

/*
 * Returns TRUE if there is max or more completions.
 */
int array_nbofcompletionshigherthan (ArrayCompletion *ac, unsigned int max)
{
    Array *a;
    unsigned int index;
    unsigned int num = 0;
    assert (ac != NULL);
    if (ac->success == COMPL_NOTDONE)
      return FALSE;
    if (ac->number != -1)
      return ac->number >= max;
    if (ac->success == COMPL_FAIL) {
        ac->number = 0;
        return FALSE;
    }
    a = ac->array;
    assert (ac->index != -1);
    index = ac->index;
    while ((index < a->nbused) &&
           (tr_strncmp(a->data[index],ac->prefix,ac->len,a->translations) == 0)) {
        index++;
        num++;
        if (num >= max)
          return TRUE;
    }
    ac->number = num;
    return FALSE;
}

/*
 * Returns the first completion of the word.
 * To be used (preferably) if there is only one completion.
 * (I can't understand its use otherwise).
 */
unsigned char *array_getcompletion (ArrayCompletion *ac)
{
    unsigned char *s;
    assert (ac != NULL);
    if (ac->success == COMPL_NOTDONE)
      return NULL;
    if (ac->success == COMPL_FAIL)
      return NULL;
    s = array_getnextcompletion (ac);
    return s;
}


/*
 * For all completions, call the function with a valid completion.
 * The word is malloced and IS NOT FREED.
 * array_forallcompletion stops as soon as forallcompletion returns TRUE.
 */

int array_forallcompletion (ArrayCompletion *ac,
                            int (*forall)(unsigned char *word, void *info),
                            void *d)
{
    unsigned char *s;
    assert (ac != NULL);
    if (ac->success == COMPL_NOTDONE)
      return ERROR;
    if (ac->success == COMPL_FAIL)
      return ERROR;
    while ((s = array_getnextcompletion (ac)) != NULL) {
        if (forall (s, d))
          return TRUE;
    }
    return FALSE;
}

/*
 * get all the completions, and put them in a array terminated with NULL.
 * nbfound also holds the number of found completions.
 * Everything is malloc'd. The user must free the data itself.
 */
unsigned char **array_getallcompletions (ArrayCompletion *ac, int *nbfound)
{
    unsigned char **r;
    unsigned char *s;
    unsigned int num;
    unsigned int i;
    if ((ac->success == COMPL_NOTDONE) || (ac->success == COMPL_FAIL))
      return NULL;
    num = array_nbofcompletions (ac);
    r = malloc ((num+1) * sizeof (char *));
    i = 0;
    while ((s = array_getnextcompletion (ac)) != NULL) {
        r[i++] = s;
    }
    r[i] = NULL;
    *nbfound = i;
    return r;
}

/*
 * return all the completion, one after the other, and return NULL when
 * there is no more word.
 * Every string is malloc'd. The user must free the data itself.
 */
unsigned char *array_getnextcompletion (ArrayCompletion *ac)
{
    unsigned char *rep;
    unsigned char *r;
    if ((ac->success == COMPL_NOTDONE) || (ac->success == COMPL_FAIL))
      return NULL;
    assert (ac->index != -1);
    if (ac->index == ac->array->nbused)
      return NULL;
    rep = ac->array->data[ac->index];
    ac->index++;
    if (tr_strncmp (rep, ac->prefix, ac->len, ac->array->translations) != 0)
      return NULL;
    r = str_dup (rep, 0);
    return r;
}



void array_dump (Array *a)
{
    unsigned int i;
    printf ("Dumping...\n");
    for (i = 0; i < a->nbused; i++) {
        printf ("%s\n", a->data[i]);
    }
    printf ("dump end\n");
}


void array_stat (Array *a)
{
    unsigned int i;
    unsigned long cumullen = 0;
    unsigned int maxlen = 0;
    printf ("Statistics:\n");
    printf ("  nb of words: %u\n", a->nbused);
    for (i = 0; i < a->nbused; i++) {
        unsigned int l = strlen (a->data[i]);
        cumullen += l;
        if (l > maxlen)
          maxlen = l;
    }
    printf ("  longest word: %u\n", maxlen);
    if (a->nbused != 0)
      printf ("  average word length: %.1f\n",
              ((double) cumullen) / a->nbused);
    if ((a->nbused != 0) && a->savable) {
        unsigned long allocatedmem;
        unsigned long strsize;
        long lostmem;
        allocatedmem = a->nbstringblocks * ALLOCPOOL;
        /* don't forget the terminating \0 => a->nbused bytes */
        strsize = cumullen + a->nbused;
        printf ("  pools used: %u (%.1f Kb)\n", a->nbstringblocks,
                allocatedmem / 1024.0);
        printf ("  total size of strings: %u (%.1f Kb)\n",
                strsize, strsize / 1024.0);
        lostmem = allocatedmem - strsize;
        printf ("  (lost mem: %d bytes, %.3f%% of total memory)\n", lostmem,
                lostmem * 100.0 / allocatedmem);
    }
}



/*
 * Save format:
 *  - the "magic" string.
 *  - the array structure (we only need nbused and nbstringblocks)
 *  - the ALLOCPOOL value (size of a pool)
 *  - the nbstringblocks "pools" (or blocks)
 *  - for every string in the array, a "bstring", that is couple
 *    of integer (index of block, offset).
 *
 * To avoid excessive write or read, the bstrings are written/read
 * in BSTRING_WRITE quantities.
 */

const char array_magic[] = "Array format ";
const char array_magic_version[] = "2.0";

#define BSTRING_WRITE 1024

struct bstring {
    int ib;                     /* index of block */
    int offset;                 /* offset inside block */
};

/* In static, so it won't take too much room in the stack. */
static struct bstring bstring_bloc[BSTRING_WRITE];

#define ctl_write(f,b,n) \
  if ((writtenbytes = write((f),(b),(n))) != (n)) {\
      if (writtenbytes == -1) perror("array_save: write"); else fprintf(stderr, "array_save: short write\n"); \
      return ERROR; }

int array_save (Array *a, int fd)
{
    unsigned int nb, ns;
    int allocpool;
    int bstring_used;
    int writtenbytes;

    if (! a->savable) {
        fprintf (stderr, "array: save: I wasn't informed you planned to save the array.\n");
        return ERROR;
    }

    if (a->nbused == 0) {
        fprintf (stderr, "array: save: cannot save an empty array.\n");
        return ERROR;
    }

    /* Ensure that the array is in a valid state. */
    array_endinsertion (a);

    ctl_write (fd, array_magic, strlen (array_magic));
    ctl_write (fd, array_magic_version, strlen (array_magic_version));
    ctl_write (fd, "\n", 2);    /* with the traling 0 */

    /* Save the array (we need only nbused and nbstringblocks) */
    ctl_write (fd, a, sizeof (Array));

    allocpool = ALLOCPOOL;
    ctl_write (fd, &allocpool, sizeof (allocpool));

    /* Save all the blocks */
    for (nb = 0; nb < a->nbstringblocks; nb++) {
        ctl_write (fd, a->blocks[nb], ALLOCPOOL);
    }

    /* And now save the array: */
    bstring_used = 0;

    /*  For every string, */
    for (ns = 0; ns < a->nbused; ns++) {
        /* find in which block it lives to find the offset */
        unsigned char *s = a->data[ns];
        int offset = -1;
        for (nb = 0; nb < a->nbstringblocks; nb++) {
            unsigned char *b = a->blocks[nb];
            /* ANSI says: one can compare two pointers only if they are
               pointers into the same array, else one invokes undefined
               behavior. But how can I know if they are in the same array
               without comparing them ? There is no way this can be written
               pedantically. */
            if ((b <= s) && (s < b + ALLOCPOOL)) {
                offset = s - b;
                break;
            }
        }
        if (offset == -1) {
            fprintf (stderr, "array: save: panic: unknown string.\n");
            return ERROR;
        }
        /* and add the information. */
        bstring_bloc[bstring_used].ib = nb;
        bstring_bloc[bstring_used].offset = offset;
        bstring_used++;
        if (bstring_used == BSTRING_WRITE) {
            ctl_write (fd, bstring_bloc,
                       BSTRING_WRITE * sizeof (struct bstring));
            bstring_used = 0;
        }
    }
    if (bstring_used != 0) {
        ctl_write (fd, bstring_bloc, bstring_used * sizeof (struct bstring));
    }

    return OK;
}

/* Remember that 'read' on anything but a file can return less bytes than what
   was asked for. (this is the case when reading from a pipe, for instance) */
static int complete_read (int f, void *buf, int n)
{
    int readbytes;
    int lu = 0;
    char *b = buf;
    while (n != 0) {
        readbytes = read (f, b, n);
        if (readbytes == -1) {
            perror ("array_load: read");
            return ERROR;
        } else if (readbytes == 0) {
            return lu;
        }
        lu += readbytes;
        n -= readbytes;
        b += readbytes;
    }
    return lu;
}

#define ctl_read(f,b,n) if (complete_read((f),(b),(n)) != (n)) return NULL

Array *array_load (int fd)
{
    Array *r;
    int allocpool;
    unsigned int nb, ns;
    int bstring_left, bstring_index;
    char magic[sizeof(array_magic)]; /* max of magic and magic_version */
    off_t off;
    char *text;

    ctl_read (fd, magic, strlen (array_magic));
    magic[strlen(array_magic)] = 0;
    if (strcmp (magic, array_magic) != 0) {
#if !HAVE_LIBZ
        fprintf (stderr, "Array: load: not an array\n");
        return NULL;
#else
        static Array *array_zload (int fd); /* further */
        lseek (fd, 0, SEEK_SET); /* rewind, */
        return array_zload (fd); /* and try again. */
#endif
    }
    ctl_read (fd, magic, strlen (array_magic_version));
    magic[strlen(array_magic_version)] = 0;
    if (strcmp (magic, array_magic_version) != 0) {
        fprintf (stderr, "Array: load: bad version %s, expecting %s.\n",
                 magic, array_magic_version);
        return NULL;
    }
    ctl_read (fd, magic, 2);    /* the last two chars */

    r =  malloc (sizeof (Array));
    if (r == NULL) {
        fprintf (stderr, "Array: load: malloc failed\n");
        return NULL;
    }
    ctl_read (fd, r, sizeof(Array));
    ctl_read (fd, &allocpool, sizeof(allocpool));

    assert (r->sorted);

    r->nballocated = r->nbused;
    r->data = malloc (r->nbused * sizeof (char *));
    if (r->data == NULL) {
        fprintf (stderr, "Array: load: malloc failed\n");
        return NULL;
    }

    r->leftinpool = 0;
    r->pool = NULL;

#if !HAVE_MMAP
    /* A loaded array is not savable, because the number of blocks is not
     * a multiple of ALLOCBUNCH (this could be easily
     * alleviated using (r->nbstringblocks/ALLOCBUNCH)*ALLOCBUNCH).
     * Moreover, if allocpool is different from ALLOCPOOL, future
     * insertion will create blocks of different size. This leads to
     * the conclusion that ALLOCPOOL should be in the array and not a
     * constant. */
    r->savable = FALSE;

    r->blocks = malloc (r->nbstringblocks * sizeof (char *));
    if (r->data == NULL) {
        fprintf (stderr, "Array: load: malloc failed\n");
        return NULL;
    }
    /* Allocate and read the "pools". */
    for (nb = 0; nb < r->nbstringblocks; nb++) {
        r->blocks[nb] = malloc (allocpool);
        if (r->blocks[nb] == NULL) {
            fprintf (stderr, "Array: load: malloc failed\n");
            return NULL;
        }
        ctl_read (fd, r->blocks[nb], allocpool);
    }
#else /* HAVE_MMAP */
    r->savable = FALSE;
    r->blocks = NULL;

    off = lseek (fd, 0, SEEK_CUR);
    text = mmap (0, allocpool * r->nbstringblocks,
                 PROT_READ, MAP_PRIVATE, fd, 0);
    if (text == (char *) -1) {
        perror ("Array: mmap");
        return NULL;
    }
    text += off;
    lseek (fd, r->nbstringblocks * allocpool, SEEK_CUR);
#endif /* HAVE_MMAP */

    /* And then the "bstrings". */
    bstring_left = 0;
    for (ns = 0; ns < r->nbused; ns++) {
        int offset;
        if (bstring_left == 0) {
            bstring_left = complete_read (fd, bstring_bloc,
                                 BSTRING_WRITE * sizeof (struct bstring));
            if (bstring_left == -1) {
                return NULL;
            }
            bstring_left /= sizeof (struct bstring);
            bstring_index = 0;
        }
        nb = bstring_bloc[bstring_index].ib;
        offset = bstring_bloc[bstring_index].offset;
#if !HAVE_MMAP
        r->data[ns] = r->blocks[nb] + offset;
#else
        r->data[ns] = text + nb*allocpool + offset;
#endif
        bstring_index++;
        bstring_left--;
    }

    /* As the array is not savable, we could free r->blocks, but as it's
       really insignificant, we won't bother.
       if (r->blocks) {free (r->blocks); r->blocks=NULL; r->nbstringblocks=0;}
    */
    return r;
}



#if HAVE_LIBZ

/*
 * This code is a copy of array_load adapted for zlib.
 * This is somewhat stupid to duplicate it, but it was becoming a true
 * mess with mmap.
 */

#define ctl_zread(f,b,n) if (gzread((f),(b),(n)) != (n)) return NULL

static Array *array_do_zload (gzFile gzf)
{
    Array *r;
    int allocpool;
    unsigned int nb, ns;
    int bstring_left, bstring_index;
    char magic[sizeof(array_magic)];

    ctl_zread (gzf, magic, strlen (array_magic));
    magic[strlen(array_magic)] = 0;
    if (strcmp (magic, array_magic) != 0) {
        fprintf (stderr, "Array: load: not an array\n");
        return NULL;
    }
    ctl_zread (gzf, magic, strlen (array_magic_version));
    magic[strlen(array_magic_version)] = 0;
    if (strcmp (magic, array_magic_version) != 0) {
        fprintf (stderr, "Array: load: bad version %s, expecting %s.\n",
                 magic, array_magic_version);
        return NULL;
    }
    ctl_zread (gzf, magic, 2);
    
    r =  malloc (sizeof (Array));
    if (r == NULL) {
        fprintf (stderr, "Array: load: malloc failed\n");
        return NULL;
    }
    ctl_zread (gzf, r, sizeof(Array));
    ctl_zread (gzf, &allocpool, sizeof(allocpool));

    assert (r->sorted);
    r->nballocated = r->nbused;
    r->data = malloc (r->nbused * sizeof (char *));
    if (r->data == NULL) {
        fprintf (stderr, "Array: load: malloc failed\n");
        return NULL;
    }

    r->leftinpool = 0;
    r->pool = NULL;
    r->savable = FALSE;
    r->blocks = malloc (r->nbstringblocks * sizeof (char *));
    if (r->data == NULL) {
        fprintf (stderr, "Array: load: malloc failed\n");
        return NULL;
    }
    for (nb = 0; nb < r->nbstringblocks; nb++) {
        r->blocks[nb] = malloc (allocpool);
        if (r->blocks[nb] == NULL) {
            fprintf (stderr, "Array: load: malloc failed\n");
            return NULL;
        }
        ctl_zread (gzf, r->blocks[nb], allocpool);
    }

    bstring_left = 0;
    for (ns = 0; ns < r->nbused; ns++) {
        int offset;
        if (bstring_left == 0) {
            bstring_left = gzread (gzf, bstring_bloc,
                                   BSTRING_WRITE * sizeof (struct bstring));
            if (bstring_left == -1) {
                return NULL;
            }
            bstring_left /= sizeof (struct bstring);
            bstring_index = 0;
        }
        nb = bstring_bloc[bstring_index].ib;
        offset = bstring_bloc[bstring_index].offset;
        r->data[ns] = r->blocks[nb] + offset;
        bstring_index++;
        bstring_left--;
    }
    return r;
}

static Array *array_zload (int fd)
{
    gzFile gzf;
    Array *r;

    gzf = gzdopen (dup (fd), "rb");
    if (gzf == NULL) {
        fprintf (stderr, "Array: load: can't do gzopen\n");
        return NULL;
    }
    r = array_do_zload (gzf);
    gzclose (gzf);
    return r;
}

#endif /* HAVE_LIBZ */

