/* $Id: array.h,v 1.7 1997/03/28 14:25:13 queinnec Exp $ */

/*
 * Alphabetic array to store words and try completions.
 *
 * Operations: inserting a word, checking its presence,
 * getting all the words starting with a given prefix (completion),
 * save and load an array.
 *
 * During a completion, one must not call endinsertion, nor remove a word.
 *
 */

#ifndef _ARRAY_H
#define _ARRAY_H

typedef struct Array Array;

/*
 * Array creation.
 * If "savable" is true, keep additional information so that array_save
 *  can work. This same information is necessary to free memory when
 *  destroying the array.
 * The translation table defines how characters are translated when comparing
 *  strings. Note that strings are inserted and handled back without change,
 *  only comparisons and sorting order use the translation table.
 * If it's NULL, a do-nothing table is used.
 */
Array *array_create (int savable, const unsigned char *translation_table);

/*
 * Array destruction.
 * If the array has not been created "savable", memory cannot be freed.
 */
int   array_destroy (Array *);

/*
 * Word insertion.
 */
int   array_insert (Array *, const unsigned char *word);

/*
 * Word removal. Rather expensive (must move up half the array).
 * No memory can be freed until array_destroy (and only if the array is "savable")
 */
int   array_delete (Array *, const unsigned char *word);

/*
 * Checking for word presence (return TRUE or FALSE).
 * The array must be sorted (via endinsertion).
 * Comparison is done according to the translation_table.
 */
int   array_check (Array *, const unsigned char *word);

/*
 * Indicates that the array can now be sorted.
 * duplicates are not removed (too expensive).
 * Not thread-safe.
 */
int   array_endinsertion (Array *);

/*
 * Return the number of words in the array.
 */
unsigned int array_nb_of_words (const Array *);

/*
 * Get the translation table.
 */
unsigned char *array_get_translations (Array *);


/* Completion */

/*
 * The array must be sorted (via endinsertion).
 * Completion is always done using the translation table.
 */

/*
 * Three way to use completions.
 *
 * The first one is:
    tc = array_createcompletion (array);
    success = array_trycompletion (tc, word);
    if (success == COMPL_FAIL) {
       echec
    } else if (success == COMPL_EXACT) {
       exact
    } else if (array_nbofcompletionshigherthan (tc, maxallowed)) {
       too many completion
    } else {
       num = array_nbofcompletions (tc);
       if (num == 1) {
          one completion
          wordcomplete = array_getcompletion (tc);
       } else {
          array_forallcompletion (tc, dosomething, info);
       }
    }
 *
 * The second way is simpler:
    tc = array_createcompletion (array);
    success = array_trycompletion (tc, word);
    while ((s = array_getnextcompletion (tc)) != NULL) {
       do anything with it.
       free (s);
    }
 *
 * The third is array_getallcompletions which returns a char** (an array of
 * all words that matches the given prefix).
 *
 * Each call to trycompletion re-initialized the ArrayCompletion.
 *
 * The three forms are mutually exclusive (more precisely, getnextcompletion,
 * forallcompletion and getallcompletions use the same fields in an
 * ArrayCompletion, and so can't be mixed).
 *
 * See each function for details.
 */

typedef struct ArrayCompletion ArrayCompletion;

#include "completion.h"

ArrayCompletion *array_createcompletion (Array *);
int             array_destroycompletion (ArrayCompletion *);

Compl_Success array_trycompletion (ArrayCompletion *, const unsigned char *word);

int array_nbofcompletions (ArrayCompletion *);

/*
 * Returns TRUE if there is max or more completions.
 */
int array_nbofcompletionshigherthan (ArrayCompletion *, unsigned int max);

/*
 * Returns the first completion of the word.
 * To be used (preferably) if there is only one completion.
 * (I can't understand its use otherwise).
 */
unsigned char *array_getcompletion (ArrayCompletion *);

/*
 * For all completions, call the function with a valid completion.
 * The word is malloced and IS NOT FREED.
 * array_forallcompletion stops as soon as forallcompletion returns TRUE.
 */
int array_forallcompletion (ArrayCompletion *,
                           int (*forall)(unsigned char* word, void* info),
                           void *);

/*
 * get all the completions, and put them in a array terminated with NULL.
 * nbfound also holds the number of found completions.
 * Everything is malloc'd. The user must free the data itself.
 */
unsigned char **array_getallcompletions (ArrayCompletion *, int *nbfound);

/*
 * return all the completions, one after the other, and return NULL when
 * there is no more word.
 * Every string is malloc'd. The user must free the data itself.
 */
unsigned char *array_getnextcompletion (ArrayCompletion *);



/* These are the first characters of the saved array. */
extern const char array_magic[];
extern const char array_magic_version[];

/*
 * Save the array.
 * The array must be marked as "savable" when it's created.
 * fd must be open for writing, of course.
 * Return OK/ERROR 
 */
int array_save (Array *array, int fd);

/*
 * Load an array.
 * The array can be modified, but it cannot be saved back again.
 * Return the new array/NULL.
 */
Array *array_load (int fd);


/* Debug */

void array_stat (Array *array);

void array_dump (Array *array);

#endif /* _ARRAY_H */
