
#include "config.h"

#include <stdio.h>

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <ctype.h>

#if HAVE_STRING_H
# include <string.h>
#else
 int strcmp (const char *, const char *);
 int strncmp (const char *, const char *, int);
 int strlen (const char *);
 char *strchr (const char *, int);
#endif

#if !HAVE_STRCHR
#  define strchr index
   char *index (const char *, int);
#endif

#include <readline/readline.h>

#include "completion.h"
#include "array.h"
#include "mycomplete.h"
#include "constants.h"
#include "exec.h"
#include "command.h"

static const char *rcsid = "$Id: command.c,v 1.18 1997/04/01 14:08:38 queinnec Exp $";

static Array *completion_array;

#ifndef TAB
#  define TAB 9 /* C-i, '\t', etc */
#endif

static unsigned char *database_complete (char *text, int state)
{
    static ArrayCompletion *ac = NULL;
    unsigned char *rep;
    if (state == 0) {
        Compl_Success success;
        ac = array_createcompletion (completion_array);
        success = array_trycompletion (ac, text);
        if (success == COMPL_FAIL) {
            array_destroycompletion (ac);
            return (NULL);
        }
    }
    rep = array_getnextcompletion (ac);
    if (rep == NULL) {
        array_destroycompletion (ac);
    }
    return (rep);
}

/*
 * This is called before doing the actual completion. It is used to set
 * the right completion function.
 * (I can't use attempted_completion_function because rl_complete first
 * initializes a lot of local variables)
 */
static void complete (int count, int key)
{
    extern Function *rl_last_func;
    int tofileorcmde = FALSE;
    int end;
    /* if we are after a > or |, use filename completion, else use
     * database completion. */
    for (end = rl_point -1 ; end != -1; end--) {
        if ((rl_line_buffer[end] == '>') || (rl_line_buffer[end] == '|')) {
            tofileorcmde = TRUE;
            break;
        }
    }
    if (tofileorcmde) {
        /* NULL means filename_completion_funtion */
        rl_completion_entry_function = (Function*) NULL;
        if (rl_last_func == (Function*) complete)
          rl_complete_internal ('?');
        else
          rl_complete_internal (TAB);
    } else {
        my_completion_translations = array_get_translations (completion_array);
        my_completion_entry_function = database_complete;
        if (rl_last_func == (Function*) complete)
          my_complete_internal ('?');
        else
          my_complete_internal (TAB);
    }
}

/*
 * Install the completion routine.
 */

static void set_completion (int docompletion, Array *array)
{
    if (docompletion)
      rl_bind_key (TAB, (Function*)complete);
    else 
      rl_bind_key (TAB, (Function*)NULL);
    completion_array = array;
}

static void help_readline (void)
{
    printf ("\nWhile entering names, you have:\n");
    printf ("  - completion, with TAB\n");
    printf ("  - full editing of current line (a la emacs)\n");
    printf ("  - history of previous entered lines (C-p, C-n and co)\n");
    printf ("\n");
}

static void init_readline (void)
{
    static int initdone = FALSE;
    if (! initdone) {
        initdone = TRUE;
        /* extern int rl_completion_query_items;
           rl_completion_query_items = 60; */
        rl_readline_name = "imoviedb";
        rl_variable_bind ("meta-flag", "on");
        rl_variable_bind ("convert-meta", "off");
        rl_variable_bind ("output-meta", "on");
        /* rl_initialize loads the inputrc file, which may hold new values
         * for convert-meta and output-meta. This is perfect for me. */
        rl_initialize();
        help_readline ();

        /* If LESSCHARSET is not set, set it to latin1, so less display
         * characters with the eighth bit set directly. This has nothing to
         * do with readline, but I don't know where to put it. */
        if (getenv ("LESSCHARSET") == NULL)
          putenv ("LESSCHARSET=latin1");
    }
}

/*
 * prompt for readline
 * quit is the string which means quit. NULL means an empty line.
 * (if line is shorter than quit but matches, this is ok)
 */
char *getline (const char *prompt, const char *quit)
{
    char *line;
    if ((line = readline (prompt)) == NULL)
      return (NULL);
    if (((*line == '\0') && (quit == NULL))
        || ((*line != '\0') && (quit != NULL) && (strncmp (quit, line, strlen (line)) == 0))) {
        free (line);
        return (NULL);
    }
    add_history (line);
    return (line);
}

/*
 * split LINE in two components: the NAME part and the OUTPUT part (aka the
 * redirection part). REDIR indicates what happens of the output.
 * REDIR is meaningless is OUTPUT is NULL.
 *
 * Accept: "name", "name>toto", "name>>toto", "name|toto", and any other
 *  variations obtained by inserting spaces.
 *
 * NAME and OUTPUT are returned as pointers into LINE.
 * LINE may be modified to suit my needs!
 */
static int split_line (char *line, char **name, char **output,
                       Redirection *redir)
{
    char *p, *q;
    
    /* find the first of '>' or '|', and set p and redir accordingly */
    p = strchr (line, '>');
    q = strchr (line, '|');
    if (p == NULL)
      p = q;
    else if ((q != NULL) && (q < p))
      p = q;
    
    if (p != NULL) {
        /* name >file or name >>file or name|cmde */
        
        q = p;
        if (*p == '|') {
            *redir = REDIR_CMDE;
        } else if (*(p+1) == '>') { /* >> redirection */
            *redir = REDIR_APPEND;
            p++;
        } else {
            *redir = REDIR_FILE;
        }

        /* remove spaces after '>' */
        for (p++; (*p != '\0') && isspace(*p); p++) /* empty */;
        if (*p == '\0') {
            fprintf (stderr, "Invalid redirection.\n");
            return (ERROR);
        } else {
            *output = p;
        }
        /* remove trailing spaces */
        p = &(line[strlen(line) - 1]);
        for ( ; isspace (*p); p--) /* empty */ ;
        *(p+1) = '\0';
        /* remove spaces before '>' */
        for (q--; (q >= line) && isspace (*q); q--) /* empty */;
        if (q == line - 1) {
            fprintf (stderr, "No name found.\n");
            return (ERROR);
        }
        *(q+1) = '\0';
        *name = line;

    } else {
        /* name only */
        
        *output = NULL;
        *name = line;
        /* remove trailing spaces */
        p = &(line[strlen(line) - 1]);
        for ( ; (p >= line) && isspace (*p); p--) /* empty */ ;
        if (p == line - 1) {
            fprintf (stderr, "No name found.\n");
            return (ERROR);
        }
        *(p+1) = '\0';
    }
    return (OK);
}



struct command {
    Array *array;
    const char *prompt;
    const char *cmde;
    const char *opt;
};

struct command the_commands[] = {
    { NULL, "name",  "list",    "-name" },
    { NULL, "name",  "ltrace",  "-name" },
    { NULL, "title", "title",   "-t" },
    { NULL, "title", "listall", "-t" },
};

struct command *command_list    = &(the_commands[0]);
struct command *command_ltrace  = &(the_commands[1]);
struct command *command_title   = &(the_commands[2]);
struct command *command_listall = &(the_commands[3]);

void init_commands (Array *namearray, Array *titlearray)
{
    the_commands[0].array = namearray;
    the_commands[1].array = namearray;
    the_commands[2].array = titlearray;
    the_commands[3].array = titlearray;
}


void name_loop (const struct command *command,
                const char *quitprompt, const char *quit)
{
    char *line;
    char *name;
    char *output;
    Redirection redir;
    init_readline ();
    set_completion (TRUE, command->array);
    while (1) {
        char prompt[100];
        sprintf (prompt, "Enter %s (%s): ", command->prompt, quitprompt);
        if ((line = getline (prompt, quit)) == NULL)
          break;
        if (split_line (line, &name, &output, &redir) == OK) {
            if (array_check (command->array, name)) {
                if (output != NULL)
                  set_command_output (redir, output);
                docommand (name, command->cmde, command->opt);
                if (output != NULL)
                  restore_default_output ();
            } else {
                printf ("Unknown name\n");
            }
        }
        free (line);
    }
}



static struct command *query_get_command ()
{
    char *line;
    int num;
    int firstcharisnum;
    set_completion (FALSE, NULL);
    while (1) {
        printf ("(1) list <name>,   (2) ltrace <name>\n");
        printf ("(3) title <title>, (4) listall <title>\n");
        printf ("(0) to quit\n");
        line = readline ("Choice: ");
        if (line == NULL)
          break;
        num = atoi (line);
        firstcharisnum = isdigit (*line);
        free (line);
        if ((!firstcharisnum) || (num < 0) || (num > 4)) {
            printf ("Invalid choice.\n");
        } else if (num == 0) {
            return (NULL);
        } else {
            return (&(the_commands[num - 1]));
        }
    }
    return (NULL);
}

void query_loop (void)
{
    init_readline ();
    while (1) {
        const struct command *command;
        command = query_get_command ();
        if (command == NULL)
          break;
        name_loop (command, "empty to go back", NULL);
    }
}
