#include "config.h"

#include <stdio.h>

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/param.h>
#ifndef MAXPATHLEN
# define MAXPATHLEN 255
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#else
 char *strcpy (char *, const char *);
 char *strchr (const char *, int);
#endif
#if !HAVE_STRCHR
#  define strchr index
   char *index (const char *, int);
#endif

#include <signal.h>

#if HAVE_VFORK_H
#  include <vfork.h>
#endif

#include "constants.h"
#include "exec.h"

static const char *rcsid = "$Id: exec.c,v 1.10 1996/10/26 16:17:56 queinnec Exp $";

/*
 * The default output parameters, set by set_default_ouput() and used by
 * restore_default_output().
 */
static int defaultoutput = REDIR_CMDE;
static char *defaultoutputname = NULL;

/* This is the current output as set by set_command_output. */
static int   output = REDIR_CMDE;
static char *outputname = NULL;

/* Possible extensions to add to a .key files. */
static char *extensions[] = { "", ".gz", ".Z", ".z", 0 };


static char *str_dup (const char *s)
{
    char *r;
    r = malloc (strlen (s) + 1);
    /* assume r != NULL */
    strcpy (r, s);
    return r;
}

/*
 * Set the default output. Should be called only once.
 * (well, you can call it more than once, but without effect).
 *
 * If name is NULL, set the default values of the default output (ouch!).
 * set_default_output (anything, NULL) can so be used to initialize the output
 *  variables with their default value.
 */
void set_default_output (Redirection redir, const char *name)
{
    if (defaultoutputname != NULL)
      return;
    if (name == NULL) {
        defaultoutput = REDIR_CMDE;
        name = (char *) getenv ("PAGER");
        if (name == NULL)
          name = MORE;
    } else {
        defaultoutput = redir;
    }
    defaultoutputname = str_dup (name);
    /* and then set the current output. */
    set_command_output (defaultoutput, defaultoutputname);
}

void restore_default_output (void)
{
    if (defaultoutputname == NULL) {
        /* PANIC!! nothing has been initialized! Do it *now*! */
        set_default_output (0, NULL);
        return;
    }
    set_command_output (defaultoutput, defaultoutputname);
}

void set_command_output (Redirection redir, const char *name)
{
    set_default_output (0, NULL); /* ensure we have been initialized */
    if (name == NULL) {
        /* aaarrrggghhh, what does this means? */
        return;
    }
    output = redir;
    if (outputname != NULL)
      free (outputname);
    outputname = str_dup (name);
}

/*
 * Execute: cmde opt w "-full"
 * The output either goes into a file or is piped into a command, based on
 *  outputtofile.
 * If outputtofile is FALSE:
 *    if outputname contains at least a space, sh -c outputname is used,
 *    else exec outputname is directly used.
 * If outputtofile is TRUE: output goes into a file.
 * This means that setting PAGER to "cat >/tmp/toto" or "tee /tmp/toto |more"
 *  works!
 */
void docommand (const char *w, const char *cmde, const char *opt)
{
    int pid;
    set_default_output (FALSE, NULL); /* ensure we have been initialized */
    signal (SIGINT, SIG_IGN);
    pid = vfork();
    if (pid == -1) {
        perror ("Cannot fork");
        return;
    } else if (pid == 0) {
        /* Child side. */
        if (output != REDIR_CMDE) {
            /* Output goes into a file. */
            int fd;
            if (output == REDIR_FILE)
              fd = open (outputname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
            else                /* REDIR_APPEND */
              fd = open (outputname, O_WRONLY|O_CREAT|O_APPEND, 0666);
            if (fd == -1) {
                perror (outputname);
                _exit (1);
            }
            close (1);
            dup (fd);
            if (execlp (cmde, cmde, opt, w, "-full", NULL) == -1) {
                perror (cmde);
                exit (1);
            }
        } else {
            /* Output goes into a command. */
            int fd[2];
            if (pipe (fd) == -1) {
                perror ("Cannot create pipe");
                _exit (1);
            }
            pid = vfork ();
            if (pid == -1) {
                perror ("Cannot fork");
                close (fd[0]);
                close (fd[1]);
                _exit (1);
            } else if (pid == 0) {
                /* Child side. Execute the command. */
                close(1);
                dup (fd[1]);
                close (fd[0]);
                if (execlp (cmde, cmde, opt, w, "-full", NULL) == -1) {
                    perror (cmde);
                    exit (1);
                }
            } else {
                /* Parent side. */
                /* (why is MORE executed in the parent side? Because the
                 * grandfather wants to waitpid() for the last son or
                 * grandson, and the MORE will end after the list/title
                 * command). */
                close (0);
                dup (fd[0]);
                close (fd[1]);
                if (strchr (outputname, ' ') != NULL) {
                    /* If there is a space in the command, I assume that it
                     * means that the command has arguments, so we need to
                     * call sh to split it up.
                     * We could almost equally use system() */
                    if (execl ("/bin/sh", "/bin/sh", "-c", outputname, NULL) == -1) {
                        perror ("/bin/sh");
                        exit (1);
                    }
                } else {
                    if (execlp (outputname, outputname, NULL) == -1) {
                        perror (outputname);
                        exit (1);
                    }
                }
            }
        }
        /* MUST NEVER BE REACHED !! */
        fprintf (stderr, "Murphy said: Unreachable code will be reached\n");
    } else {
        /* Parent side. */
#if HAVE_WAITPID
        waitpid (pid, NULL, 0);
#else
        int child;
        do {
            child = wait (NULL);
            if (child == -1)
              break; /* To be honest, it's an error. */
        } while (child != pid);
#endif
    signal (SIGINT, SIG_DFL);
    }
}

#ifdef notdef /*COMPRESS*/
/* The code has been disabled because what I am looking for is speed,
   and uncompressing the file is taking more than five time longer as
   the load part. Moreover, it won't work with mmap.
   On the other hand, the zlibrary is quite fast, but one still loses mmap.
   I keep it here, because it's a nice way of doing it. */
int openfile_readonly (const char *filename)
{
    int fd;
    char zfilename[MAXPATHLEN];
    int p[2];
    int pid;
    if ((fd = open (filename, O_RDONLY)) != -1)
      return fd;
    sprintf (zfilename, "%s%s", filename, ZDBSEXT);
    fd = open (zfilename, O_RDONLY);
    if (fd == -1)
      return -1;
    if (pipe (p) == -1) {
        perror ("cannot pipe");
        close (fd);
        return -1;
    }
    pid = vfork ();
    if (pid == -1) {
        perror ("cannot fork");
        close (fd); close (p[0]); close (p[1]);
        return -1;
    } else if (pid == 0) { /* child */
        char *cmde = ZDBSCAT;
        char *opt = ZDBSCATOPTS;
        close (p[0]);
        if (dup2 (p[1], 1) == -1) perror ("dup2");
        if (dup2 (fd, 0) == -1) perror ("dup2");
        close (fd);
        if ((strchr (cmde, ' ') != NULL) || (strchr (opt, ' ') != NULL)) {
            char fullcmde[MAXPATHLEN + 10];
            sprintf (fullcmde, "%s %s", cmde, opt);
            execlp ("/bin/sh", "/bin/sh", "-c", fullcmde, NULL);
        } else if (opt[0] != 0) {
            execlp (cmde, cmde, opt, NULL);
        } else {
            execlp (cmde, cmde, NULL);
        }
        perror ("can't execlp");
    } else { /* parent */
        close (p[1]);
        close (fd);
        return p[0];
    }
}
#endif

int openfile_readonly (const char *filename)
{
    char zfilename[MAXPATHLEN];
    char **ext;
    int fd;
    for (ext = extensions; *ext != NULL; ext++) {
        sprintf (zfilename, "%s%s", filename, *ext);
        fd = open (zfilename, O_RDONLY);
        if (fd != -1)
          return fd;
    }
    return -1;
}


int opencompletion (const char *mainfile, const char *checkfile)
{
    int fd;
    struct stat mainstat;
    struct stat checkstat;
    char **ext;

    fd = openfile_readonly (mainfile);
    if (fd == -1)
      return -1;

    if (fstat (fd, &mainstat) == -1) {
        fprintf (stderr, "Can't stat fd of %s: ", mainfile);
        perror ("");
        return fd;
    }

    for (ext = extensions; *ext != NULL; ext++) {
        char zfilename[MAXPATHLEN];
        sprintf (zfilename, "%s%s", checkfile, *ext);
        if (stat (zfilename, &checkstat) != -1)
          break;
    }
    if (*ext == NULL)           /* checkfile not found, too bad */
      return fd;

    if (mainstat.st_mtime < checkstat.st_mtime) {
        fprintf (stderr,
                 "Warning: %s is older than %s. cgencompl should be run.\n",
                 mainfile, checkfile);
    }
    return fd;
}
