/* complete.c -- filename completion for readline. */
/* Modified by P. Queinnec <queinnec@enseeiht.fr>
   Simplified for imoviedb:
    - remove everything about shells, filename, users...
    - remove word breaking code and variables
    - completion_matches() uses a translation_array.
    */

/* Copyright (C) 1987, 1989, 1992 Free Software Foundation, Inc.

   This file is part of the GNU Readline Library, a library for
   reading lines of text with interactive input and history editing.

   The GNU Readline Library is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 1, or
   (at your option) any later version.

   The GNU Readline Library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   675 Mass Ave, Cambridge, MA 02139, USA. */

#include "config.h"

#include <stdio.h>

#if defined (HAVE_STDLIB_H)
#  include <stdlib.h>
#else
extern void abort();
extern void qsort();
#endif /* HAVE_STDLIB_H */

/* Some standard library routines. */
#include <readline/readline.h>

#include "translation.h"

static const char *rcsid = "$Id: mycomplete.c,v 1.3 1997/03/29 14:36:02 queinnec Exp $";


extern char *rl_copy_text ();

/* From readline library */
extern int screenwidth;

/* Forward declarations for functions defined and used in this file. */
static unsigned char **my_completion_matches (char *text,
                                              unsigned char *(*entry_function)());
static int compare_strings (const void *, const void *);

extern char *xmalloc (), *xrealloc ();




/* **************************************************************** */
/*								    */
/*	Completion matching, from readline's point of view.	    */
/*								    */
/* **************************************************************** */

/* The translation table.
   Must be a non-NULL char[256]. */
unsigned char *my_completion_translations = NULL;

/* Pointer to the generator function for completion_matches (). */
unsigned char *(*my_completion_entry_function)() = NULL;

/* The user must press "y" or "n". Non-zero return means "y" pressed. */
static int get_y_or_n ()
{
  int c;

  for (;;)
    {
      c = rl_read_key ();
      if (c == 'y' || c == 'Y' || c == ' ')
	return (1);
      if (c == 'n' || c == 'N' || c == RUBOUT)
	return (0);
      if (c == ABORT_CHAR)
	rl_abort ();
      ding ();
    }
}

/* Up to this many items will be displayed in response to a
   possible-completions call.  After that, we ask the user if
   she is sure she wants to see them all. */
extern int rl_completion_query_items; /* = 100; */


/* Complete the word at or before point.
   WHAT_TO_DO says what to do with the completion.
   `?' means list the possible completions.
   TAB means do standard completion. */
int my_complete_internal (what_to_do)
     int what_to_do;
{
  unsigned char **matches;
  int start, end;
  unsigned char *text;
  unsigned char *replacement;

  start = 0;
  end = rl_point;
  text = rl_copy_text (start, end);

  matches = my_completion_matches (text, my_completion_entry_function);

  free (text);

  if (!matches)
    ding ();
  else
    {
      register int i;

      unsigned char *lowest_common;
      int j, newlen = 0;
      unsigned char dead_slot;
      unsigned char **temp_array;
      
      /* Sort the items. */
      /* It is safe to sort this array, because the lowest common
         denominator found in matches[0] will remain in place. */
      for (i = 0; matches[i]; i++);
      qsort (matches, i, sizeof (char *), compare_strings);
      
      /* Remember the lowest common denominator for it may be unique. */
      lowest_common = savestring (matches[0]);
      
      for (i = 0; matches[i + 1]; i++)
          {
	      if (tr_strcmp (matches[i], matches[i + 1], my_completion_translations) == 0)
                  {
                      free (matches[i]);
                      matches[i] = &dead_slot;
                  }
	      else
		newlen++;
          }
      
      /* We have marked all the dead slots with (char *)&dead_slot.
         Copy all the non-dead entries into a new array. */
      temp_array = (unsigned char **)xmalloc ((3 + newlen) * sizeof (char *));
      for (i = j = 1; matches[i]; i++)
          {
	      if (matches[i] != &dead_slot)
		temp_array[j++] = matches[i];
          }
      temp_array[j] = NULL;
      
      if (matches[0] != &dead_slot)
        free (matches[0]);
      free (matches);

      matches = temp_array;
      
      /* Place the lowest common denominator back in [0]. */
      matches[0] = lowest_common;

      /* If there is one string left, and it is identical to the
         lowest common denominator, then the LCD is the string to
         insert. */
      if (j == 2 && tr_strcmp (matches[0], matches[1], my_completion_translations) == 0)
          {
	      free (matches[1]);
	      matches[1] = NULL;
          }

      switch (what_to_do)
	{
	case TAB:
	  replacement = matches[0];

	  if (replacement)
	    {
	      rl_begin_undo_group ();
	      rl_delete_text (start, rl_point);
	      rl_point = start;
	      rl_insert_text (replacement);
	      rl_end_undo_group ();
	    }

	  /* If there are more matches, ring the bell to indicate. */
	  if (matches[1])
	    {
		ding ();	/* There are other matches remaining. */
	    }
	  break;

	case '?':
	  {
	    int len, count, limit, max;
	    int j, k, l;

	    /* Handle simple case first.  What if there is only one answer? */
	    if (!matches[1])
	      {
		crlf ();
		fputs (matches[0], rl_outstream);
		crlf ();
		goto restart;
	      }

	    /* There is more than one answer.  Find out how many there are,
	       and find out what the maximum printed length of a single entry
	       is. */
	    for (max = 0, i = 1; matches[i]; i++)
	      {
		unsigned char *temp;
		int name_length;

		temp = matches[i];
		name_length = strlen (temp);

		if (name_length > max)
		  max = name_length;
	      }

	    len = i - 1;

	    /* If there are many items, then ask the user if she
	       really wants to see them all. */
	    if (len >= rl_completion_query_items)
	      {
		crlf ();
		fprintf (rl_outstream,
			 "There are %d possibilities.  Do you really", len);
		crlf ();
		fprintf (rl_outstream, "wish to see them all? (y or n)");
		fflush (rl_outstream);
		if (!get_y_or_n ())
		  {
		    crlf ();
		    goto restart;
		  }
	      }

	    /* How many items of MAX length can we fit in the screen window? */
	    max += 2;
	    limit = screenwidth / max;
	    if (limit != 1 && (limit * max == screenwidth))
	      limit--;

	    /* Avoid a possible floating exception.  If max > screenwidth,
	       limit will be 0 and a divide-by-zero fault will result. */
	    if (limit == 0)
	      limit = 1;

	    /* How many iterations of the printing loop? */
	    count = (len + (limit - 1)) / limit;

	    /* Watch out for special case.  If LEN is less than LIMIT, then
	       just do the inner printing loop. */
	    if (len < limit)
	      count = 1;

	    /* Print the sorted items, up-and-down alphabetically, like
	       ls might. */
	    crlf ();

	    for (i = 1; i < count + 1; i++)
	      {
		for (j = 0, l = i; j < limit; j++)
		  {
		    if (l > len || !matches[l])
		      break;
		    else
		      {
			int printed_length;

			printed_length = strlen (matches[l]);
                        fputs (matches[l], rl_outstream);

			if (j + 1 < limit)
			  {
			    for (k = 0; k < max - printed_length; k++)
			      putc (' ', rl_outstream);
			  }
		      }
		    l += count;
		  }
		crlf ();
	      }
	  restart:

	    rl_on_new_line ();
	  }
	  break;

	default:
	  fprintf (stderr, "\r\nreadline: bad value for what_to_do in rl_complete\n");
	  abort ();
	}

      for (i = 0; matches[i]; i++)
	free (matches[i]);
      free (matches);
    }

  return 0;
}

#define tr(c) my_completion_translations[c]

/* Stupid comparison routine for qsort () ing strings. */
static int
compare_strings (const void *p1, const void *p2)
{
  unsigned char **s1 = (unsigned char **)p1;
  unsigned char **s2 = (unsigned char **)p2;
  int result;

  result = tr(**s1) - tr(**s2);
  if (result == 0)
    result = tr_strcmp (*s1, *s2, my_completion_translations);

  return result;
}


/* **************************************************************** */
/*								    */
/*			     Completion				    */
/*								    */
/* **************************************************************** */

/* Return an array of (char *) which is a list of completions for TEXT.
   If there are no completions, return a NULL pointer.
   The first entry in the returned array is the substitution for TEXT.
   The remaining entries are the possible completions.
   The array is terminated with a NULL pointer.

   ENTRY_FUNCTION is a function of two args, and returns a (char *).
     The first argument is TEXT.
     The second is a state argument; it should be zero on the first call, and
     non-zero on subsequent calls.  It returns a NULL pointer to the caller
     when there are no more matches.
 */
static unsigned char **
my_completion_matches (char *text, unsigned char *(*entry_function)())
{
  /* Number of slots in match_list. */
  int match_list_size;

  /* The list of matches. */
  unsigned char **match_list =
    (unsigned char **)xmalloc (((match_list_size = 10) + 1) * sizeof (char *));

  /* Number of matches actually found. */
  int matches = 0;

  /* Temporary string binder. */
  unsigned char *string;

  match_list[1] = (char *)NULL;

  while ((string = (*entry_function) (text, matches)))
    {
      if (matches + 1 == match_list_size)
	match_list = (unsigned char **)xrealloc
	  (match_list, ((match_list_size += 10) + 1) * sizeof (char *));

      match_list[++matches] = string;
      match_list[matches + 1] = (char *)NULL;
    }

  /* If there were any matches, then look through them finding out the
     lowest common denominator.  That then becomes match_list[0]. */
  if (matches)
    {
      register int i = 1;
      int low = 100000;		/* Count of max-matched characters. */

      /* If only one match, just use that. */
      if (matches == 1)
	{
	  match_list[0] = match_list[1];
	  match_list[1] = (char *)NULL;
	}
      else
	{
	  /* Otherwise, compare each member of the list with
	     the next, finding out where they stop matching. */
          while (i < matches)
	    {
	      register int si;
              unsigned char c1, c2;

              for (si = 0;
                   (c1 = match_list[i][si]) &&
                     (c2 = match_list[i + 1][si]);
                   si++)
                if (tr(c1) != tr(c2)) break;

	      if (low > si) low = si;
	      i++;
	    }
	  match_list[0] = xmalloc (low + 1);
	  strncpy (match_list[0], match_list[1], low);
	  match_list[0][low] = '\0';
	}
    }
  else				/* There were no matches. */
    {
      free (match_list);
      match_list = NULL;
    }

  return (match_list);
}


