#include "config.h"

#include <stdio.h>

#if HAVE_STRING_H
#include <string.h>
#else
 char *strchr (const char *, int);
#endif
#if !HAVE_STRCHR
#  define strchr index
   char *index (const char *, int);
#endif

#include <moviedb.h>
#include <dbutils.h>

#include "array.h"
#include "translation.h"
#include "constants.h"
#include "readname.h"
#include "exec.h"

static const char *rcsid = "$Id: readname.c,v 1.20 1997/06/27 09:00:28 queinnec Exp $";

/*
 * Read filename (titles.key format) into array.
 * If stripyear is true, remove year information /(1/, /(?/ or /( /
 * If fromyear is not 0, keep only movies made after fromyear (inclusive)
 * If toyear is not 0, keep only movies made before toyear (exclusive)
 * (See Dijkstra, "Why numbering should start at Zero", in "Fillers at the
 * YoP Institute", in "Formal Development of Programs and Proofs", edited
 * by Dijkstra)
 */
int readname_titlekeyfile (Array *array, const char *filename, int stripyear,
                           unsigned int fromyear, unsigned int toyear)
{
    FILE *f;
    char line[MXLINELEN];
    char prevline[MXLINELEN];
    char *p;
    unsigned int nbinserted;
    unsigned int linenum = 0;
    int needyear = stripyear || (fromyear != 0) || (toyear != 0);
    
    f = openFile (filename);
    if (f == NULL) {
        fprintf (stderr, "Can't open %s: ", filename);
        perror ("");
        return (ERROR);
    }
    fprintf (stderr, "Loading %s.", filename);
    prevline[0] = 0;
    while (1) {
        linenum++;
        if (linenum % 1000 == 0)
          fprintf (stderr, ".");
        (void) fgets (line, MXLINELEN, f) ;
        if  ((p = strchr (line, FSEP)) == NULL)
          break;
        if (p == line)          /* bogus line */
          continue;
        *p = 0;

        if (needyear) {
            for (p = line; (p = strchr (p, '(')) != NULL; p++) {
                if ((*(p+1) == '1') || (*(p+1) == '2')
                    || (*(p+1) == '?') || (*(p+1) == ' ')) {
                    /* year information has probably been found */
                    if (p != line)
                      /* first char could be '('! (no such film currently) */
                      break;
                }
            }
            
            if (p != NULL) {    /* year found */
                /* p == "(year)" */
                if ((fromyear != 0) || (toyear != 0)) {
                    unsigned int theyear;
                    if (*(p+1) == '?') {
                        theyear = 0;
                    } else {
                        if (sscanf (p, "(%u)", &theyear) != 1) {
                            fprintf (stderr, "readname panic: malformed year in <%s>\n", line);
                            continue; /* next line */
                        }
                    }
                    if ((fromyear != 0) && ((theyear == 0) || (theyear < fromyear))) {
                        continue; /* title ignored, read next line */
                    }
                    if ((toyear != 0) && ((theyear == 0) || (theyear >= toyear))) {
                        continue; /* title ignored, read next line */
                    }
                }
                if (stripyear) {
                    if ((p > line) && (*(p-1) == ' ')) {
                        /* remove space before '(' */
                        p--;
                    }
                    *p = '\0';
                }
            } /* if year found */
        } /* if needyear */

        if (stripyear) {
            if (strcmp (line, prevline) == 0) {
                /* ignore duplicate title */
                continue;
            }
            strcpy (prevline, line);
        }
        
        if (array_insert (array, line) == ERROR)
          return (ERROR);
    }
    
    fclose (f);
    fprintf (stderr, "sorting...");
    array_endinsertion (array);
    nbinserted = array_nb_of_words (array);
    fprintf (stderr, "%u names read, %u inserted.\n", linenum, nbinserted);
    return (OK);
}

int readname_namekeyfile (Array *array, const char *filename)
{
    FILE *f;
    char line[MXLINELEN];
    char *p;
    unsigned int nbinserted;
    unsigned int linenum = 0;
    
    f = openFile (filename);
    if (f == NULL) {
        fprintf (stderr, "Can't open %s: ", filename);
        perror ("");
        return (ERROR);
    }
    fprintf (stderr, "Loading %s.", filename);
    while (1) {
        (void) fgets (line, MXLINELEN, f) ;
        if  ((p = strchr (line, FSEP)) == NULL)
          break;
        if (p == line)          /* bogus line */
          continue;
        *p = 0;
        if (array_insert (array, line) == ERROR)
          return (ERROR);
        linenum++;
        if (linenum % 1000 == 0)
          fprintf (stderr, ".");
    }
    fclose (f);
    fprintf (stderr, "sorting...");
    array_endinsertion (array);
    nbinserted = array_nb_of_words (array);
    fprintf (stderr, "%u names read, %u inserted.\n", linenum, nbinserted);
    return (OK);
}

int readname_dbs (Array *array, const char *dbsstem)
{
    FILE *indexfp;
    FILE *namekeyfp;
    FILE *nameidxfp;
    char dbs[MAXPATHLEN];
    long namekey;
    unsigned int linenum = 0;
    const char *name;

    sprintf (dbs, "%s.%s", dbsstem, NDXEXT);
    indexfp = openFile (dbs);
    namekeyfp = openFile (NAMEKEY);
    nameidxfp = openFile (NAMEIDX);
    fprintf (stderr, "Loading %s.", dbs);
    while (! feof(indexfp)) {
        namekey = getName (indexfp);
        if (namekey == -1)
          break;
        (void) getOffset (indexfp) ;
        name = mapfastNameKeyToText (namekey, namekeyfp, nameidxfp);
        if (array_insert (array, name) == ERROR)
          return (ERROR);
        linenum++;
        if (linenum % 1000 == 0)
          fprintf (stderr, ".");
    }
    fprintf (stderr, "%u names.\n", linenum);
    fclose (indexfp);
    fclose (namekeyfp);
    fclose (nameidxfp);
    return (OK);
}

Array *readname_loadtitle (const char *complfile,
                           const char *keyfile,
                           int stripyear,
                           unsigned int fromyear, unsigned int toyear)
{
    Array *thearray;
    int fd;
    if ((stripyear) && (fromyear == 0) && (toyear == 0) &&
        (complfile != NULL) && ((fd = opencompletion (complfile, keyfile)) != -1)) {
        fprintf (stderr, "Restoring state from %s ", complfile);
        thearray = array_load (fd);
        close (fd);
        if (thearray == NULL) {
            fprintf (stderr, "restoration failed.\n");
            exit (1);
        }
        fprintf (stderr, "%u titles.\n", array_nb_of_words (thearray));
    } else {
        thearray = array_create (TRUE, translations_iso2ascii_lc);
        if (thearray == NULL)
          exit (1);
        if (readname_titlekeyfile (thearray, keyfile, stripyear, fromyear, toyear) == ERROR) {
            fprintf (stderr, "%s not loaded\n", keyfile);
            exit (1);
        }
    }
    return (thearray);
}

Array *readname_loadname (const char *complfile, const char *keyfile)
{
    Array *thearray;
    int fd;
    if ((fd = opencompletion (complfile, keyfile)) != -1) {
        fprintf (stderr, "Restoring state from %s ", complfile);
        thearray = array_load (fd);
        close (fd);
        if (thearray == NULL) {
            fprintf (stderr, "restoration failed.\n");
            exit (1);
        }
        fprintf (stderr, "%u names.\n", array_nb_of_words (thearray));
    } else {
        thearray = array_create (TRUE, translations_iso2ascii_lc);
        if (thearray == NULL)
          exit (1);
        if (readname_namekeyfile (thearray, keyfile) == ERROR) {
            fprintf (stderr, "%s not loaded\n", keyfile);
            exit (1);
        }
    }
    return thearray;
}
